import hashlib

from hmmer import HMMER, example_filepath


def test_fetch():
    hmmfile = example_filepath("Pfam-A_24.hmm")
    hmm = HMMER(hmmfile)
    profiles = hmm.fetch(["PF01797.17", "PF03447.17"])

    hex = hashlib.sha256(profiles.encode()).hexdigest()
    assert hex == "f9ac45bb2cea5c0d610a6fe6f3a707246b4cc69926d66f07d67bd1346be06ded"



import csv
from pathlib import Path
from typing import IO, List, NamedTuple, Union

from ._misc import decomment

__all__ = ["TBLScore", "TBLRow", "TBLIndex", "TBLDom", "read_tbl"]

TBLIndex = NamedTuple("TBLIndex", [("name", str), ("accession", str)])

TBLScore = NamedTuple("TBLScore", [("e_value", str), ("score", str), ("bias", str)])

TBLDom = NamedTuple(
    "TBLDom",
    [
        ("exp", str),
        ("reg", int),
        ("clu", int),
        ("ov", int),
        ("env", int),
        ("dom", int),
        ("rep", int),
        ("inc", int),
    ],
)

TBLRow = NamedTuple(
    "TBLRow",
    [
        ("target", TBLIndex),
        ("query", TBLIndex),
        ("full_sequence", TBLScore),
        ("best_1_domain", TBLScore),
        ("domain_numbers", TBLDom),
        ("description", str),
    ],
)


def read_tbl(file: Union[str, Path, IO[str]]) -> List[TBLRow]:
    """
    Read tbl file type.

    Parameters
    ----------
    file
        File path or file stream.
    """
    closeit = False
    if isinstance(file, str):
        file = Path(file)

    if isinstance(file, Path):
        file = open(file, "r")
        closeit = True

    rows = []
    for line in csv.reader(decomment(file), delimiter=" ", skipinitialspace=True):
        row = TBLRow(
            TBLIndex(line[0], line[1]),
            TBLIndex(line[2], line[3]),
            TBLScore(*line[4:7]),
            TBLScore(*line[7:10]),
            TBLDom(line[10], *[int(i) for i in line[11:18]]),
            " ".join(line[18:]),
        )
        rows.append(row)

    if closeit:
        file.close()

    return rows



import logging
from pathlib import Path

import pooch

__all__ = ["example_filepath"]

pooch.get_logger().setLevel(logging.ERROR)

goodboy = pooch.create(
    path=pooch.os_cache("hmmer"),
    base_url="https://hmmer-py.s3.eu-west-2.amazonaws.com/",
    registry={
        "Pfam-A_24.hmm.gz": "32791a1b50837cbe1fca1376a3e1c45bc84b32dd4fe28c92fd276f3f2c3a15e3",
    },
)


def example_filepath(filename: str) -> Path:
    return Path(goodboy.fetch(filename + ".gz", processor=pooch.Decompress()))



import tempfile
from enum import Enum
from io import StringIO
from pathlib import Path
from subprocess import PIPE, Popen, check_call, check_output
from typing import List, Optional, TextIO, Union

from fasta_reader import FASTAItem, read_fasta

from ._misc import make_path
from .bin import hmmemit, hmmfetch, hmmpress, hmmscan, hmmsearch
from .domtbl import DomTBLRow, read_domtbl
from .tbl import TBLRow, read_tbl

__all__ = ["HMMER", "Result"]


class State(Enum):
    UNKNOWN = 1
    YES = 2
    NO = 3


class Result:
    def __init__(self, tbl: Optional[Path] = None, domtbl: Optional[Path] = None):
        if tbl is None:
            self._tbl = None
        else:
            self._tbl = read_tbl(tbl)

        if domtbl is None:
            self._domtbl = None
        else:
            self._domtbl = read_domtbl(domtbl)

    @property
    def has_tbl(self) -> bool:
        return self._tbl is not None

    @property
    def has_domtbl(self) -> bool:
        return self._domtbl is not None

    @property
    def tbl(self) -> List[TBLRow]:
        assert self._tbl is not None
        return self._tbl

    @property
    def domtbl(self) -> List[DomTBLRow]:
        assert self._domtbl is not None
        return self._domtbl


def _optional_filepath(
    filepath_or_bool: Union[Path, str, bool], tmp_filepath: Path
) -> Optional[Path]:

    if isinstance(filepath_or_bool, str):
        filepath_or_bool = Path(filepath_or_bool)

    opt_filepath: Optional[Path] = None
    if isinstance(filepath_or_bool, bool):
        if filepath_or_bool:
            opt_filepath = tmp_filepath
    else:
        opt_filepath = filepath_or_bool

    return opt_filepath


class Options:
    def __init__(
        self,
        output: Optional[Union[Path, str]],
        tblout: Optional[Path],
        domtblout: Optional[Path],
        heuristic: bool,
        cut_ga: bool,
        hmmkey: Optional[str],
        Z: Optional[int],
    ):
        self._options = []

        if output is not None:
            self._options += ["-o", str(output)]

        if tblout is not None:
            self._options += ["--tblout", str(tblout)]

        if domtblout is not None:
            self._options += ["--domtblout", str(domtblout)]

        if not heuristic:
            self._options += ["--max"]

        if cut_ga:
            self._options += ["--cut_ga"]

        if Z:
            self._options += ["-Z", str(Z)]

        self._tblout = tblout
        self._domtblout = domtblout
        self._hmmkey = hmmkey

    def aslist(self):
        return self._options

    @property
    def tblout(self):
        return self._tblout

    @property
    def domtblout(self):
        return self._domtblout

    @property
    def has_hmmkey(self) -> bool:
        return self._hmmkey is not None

    @property
    def hmmkey(self) -> str:
        assert isinstance(self._hmmkey, str)
        return self._hmmkey


def make_target(target: Union[Path, str, TextIO], tmpdir: Path) -> Path:
    if isinstance(target, str):
        target = Path(target)

    if isinstance(target, Path):
        return target

    with open(tmpdir / "target.fasta", "w") as file:
        file.write(target.read())

    return tmpdir / "target.fasta"


class HMMER:
    def __init__(self, profile: Union[Path, str]):
        self._profile = make_path(profile).absolute()
        self._indexed = State.UNKNOWN
        self._timeout = 15

    @property
    def timeout(self) -> int:
        return self._timeout

    @timeout.setter
    def timeout(self, timeout: int):
        self._timeout = timeout

    def index(self):
        check_call([str(hmmfetch), "--index", self._profile])

    @property
    def is_indexed(self) -> bool:
        if self._indexed == State.UNKNOWN:
            prof = self._profile
            if prof.with_suffix(prof.suffix + ".ssi").exists():
                self._indexed = State.YES
            elif prof.with_suffix(prof.suffix + ".h3m.ssi").exists():
                self._indexed = State.YES
            else:
                self._indexed = State.NO

        return self._indexed == State.YES

    def fetch_profile(self, key: str) -> str:
        return check_output([str(hmmfetch), str(self._profile), key], text=True)

    def press(self):
        check_call([hmmpress, self._profile])

    def fetch(self, profile_accs: List[str]) -> str:
        with tempfile.TemporaryDirectory() as tmpdir:
            keyfile = Path(tmpdir) / "keyfile"
            with open(keyfile, "w") as file:
                for acc in profile_accs:
                    file.write(f"{acc}\n")

            cmd = f"{hmmfetch} -f {self._profile} {keyfile}"
            return check_output(cmd, shell=True, text=True)

    @property
    def is_pressed(self) -> bool:
        p = self._profile
        exts = [".h3f", ".h3i", ".h3m", ".h3p"]
        return all([p.with_suffix(p.suffix + ext).exists() for ext in exts])

    def emit(self, hmmkey: str, nsamples=1, consensus=False, seed=0) -> List[FASTAItem]:
        options = []
        if consensus:
            options.append("-c")
        if nsamples != 1:
            options += ["-N", str(nsamples)]
        if seed != 0:
            options += ["-seed", str(seed)]

        flags = " ".join(options)
        cmd = f"{hmmfetch} {self._profile} {hmmkey} | {hmmemit} {flags} -"
        output = check_output(cmd, shell=True, text=True)
        return read_fasta(StringIO(output)).read_items()

    def scan(
        self,
        target: Union[Path, str, TextIO],
        output: Optional[Union[Path, str]] = None,
        tblout: Union[Path, str, bool] = True,
        domtblout: Union[Path, str, bool] = True,
        heuristic=True,
        cut_ga=False,
        hmmkey: Optional[str] = None,
        Z: Optional[int] = None,
    ) -> Result:

        with tempfile.TemporaryDirectory() as tmpdir:
            tbl_file = _optional_filepath(tblout, Path(tmpdir) / "tbl.txt")
            domtbl_file = _optional_filepath(domtblout, Path(tmpdir) / "domtbl.txt")

            opts = Options(output, tbl_file, domtbl_file, heuristic, cut_ga, hmmkey, Z)
            target = make_target(target, Path(tmpdir))
            return self._match(hmmscan, target, opts)

    def search(
        self,
        target: Union[Path, str, TextIO],
        output: Optional[Union[Path, str]] = None,
        tblout: Union[Path, str, bool] = True,
        domtblout: Union[Path, str, bool] = True,
        heuristic=True,
        cut_ga=False,
        hmmkey: Optional[str] = None,
        Z: Optional[int] = None,
    ) -> Result:

        with tempfile.TemporaryDirectory() as tmpdir:
            tbl_file = _optional_filepath(tblout, Path(tmpdir) / "tbl.txt")
            domtbl_file = _optional_filepath(domtblout, Path(tmpdir) / "domtbl.txt")

            opts = Options(output, tbl_file, domtbl_file, heuristic, cut_ga, hmmkey, Z)
            target = make_target(target, Path(tmpdir))
            return self._match(hmmsearch, target, opts)

    def _match(self, bin: Path, target: Path, options: Options) -> Result:
        target = target.absolute()

        cmd_match = [str(bin)] + options.aslist()

        if options.has_hmmkey:

            cmd_fetch = [str(hmmfetch), str(self._profile), options.hmmkey]
            with Popen(cmd_fetch, stdout=PIPE) as pfetch:
                cmd_match += ["-", str(target)]
                with Popen(cmd_match, stdin=pfetch.stdout) as pmatch:
                    pmatch.wait(timeout=self._timeout)

        else:
            cmd_match += [str(self._profile), str(target)]
            check_call(cmd_match)

        return Result(options.tblout, options.domtblout)



from io import StringIO

import pytest

from hmmer import read_domtbl, read_tbl

_tbl_content = """#                                                               --- full sequence ---- --- best 1 domain ---- --- domain number estimation ----
# target name        accession  query name           accession    E-value  score  bias   E-value  score  bias   exp reg clu  ov env dom rep inc description of target
#------------------- ---------- -------------------- ---------- --------- ------ ----- --------- ------ -----   --- --- --- --- --- --- --- --- ---------------------
item2                -          Octapeptide          PF03373.14   1.2e-07   19.5   3.5   1.2e-07   19.5   3.5   1.0   1   0   0   1   1   1   1 Description one two three
item3                -          Octapeptide          PF03373.14   1.2e-07   19.5   3.5   1.2e-07   19.5   3.5   1.0   1   0   0   1   1   1   1 -
#
# Program:         hmmsearch
# Version:         3.2.1 (June 2018)
# Pipeline mode:   SEARCH
# Query file:      tmp/database.hmm
# Target file:     tmp/amino.fasta
# Option settings: hmmsearch -A tmp/db_hits --tblout tmp/db_tblout tmp/database.hmm tmp/amino.fasta
# Current dir:     /Users/horta/code/nmm-py
# Date:            Fri Jan 10 12:41:26 2020
# [ok]
"""

_domtbl_content = """#                                                                                   --- full sequence --- -------------- this domain -------------   hmm coord   ali coord   env coord
# target name        accession   tlen query name                  accession   qlen   E-value  score  bias   #  of  c-Evalue  i-Evalue  score  bias  from    to  from    to  from    to  acc description of target
#------------------- ---------- -----        -------------------- ---------- ----- --------- ------ ----- --- --- --------- --------- ------ ----- ----- ----- ----- ----- ----- ----- ---- ---------------------
Leader_Thr           PF08254.12    22 AE014075.1:190-252|amino|11 -             21   5.3e-10   38.8  17.5   1   1     3e-14   5.5e-10   38.8  17.5     1    22     1    21     1    21 0.99 Threonine leader peptide
Y1_Tnp               PF01797.17   121 AE014075.1:534-908|amino|11 -            125   8.3e-25   87.2   0.0   1   1     5e-29   9.2e-25   87.1   0.0    19   121     2   102     1   102 0.94 Transposase IS200 like
Homoserine_dh        PF00742.20   173 AE014075.1:985-3507|amino|11 -            841   1.5e-49  168.2   0.0   1   1     2e-52   7.3e-49  166.0   0.0     1   173   635   832   635   832 0.92 Homoserine dehydrogenase
AA_kinase            PF00696.29   241 AE014075.1:985-3507|amino|11 -            841   1.9e-46  158.7   0.3   1   1   1.6e-49   5.7e-46  157.1   0.1     3   241    23   305    21   305 0.87 Amino acid kinase family
NAD_binding_3        PF03447.17   117 AE014075.1:985-3507|amino|11 -            841   5.9e-34  117.3   0.0   1   1   4.6e-37   1.7e-33  115.8   0.0     1   116   493   626   493   627 0.97 Homoserine dehydrogenase, NAD binding domain
ACT_7                PF13840.7     65 AE014075.1:985-3507|amino|11 -            841   4.5e-19   68.0   1.3   1   2   4.7e-11   1.7e-07   30.9   0.0     5    63   335   395   333   397 0.90 ACT domain
ACT_7                PF13840.7     65 AE014075.1:985-3507|amino|11 -            841   4.5e-19   68.0   1.3   2   2     2e-12   7.2e-09   35.3   0.3     3    61   414   474   412   478 0.92 ACT domain
ACT                  PF01842.26    67 AE014075.1:985-3507|amino|11 -            841   1.8e-17   62.8   0.3   1   2   2.3e-11   8.2e-08   31.9   0.0     6    61   345   397   341   402 0.91 ACT domain
#
# Program:         hmmscan
# Version:         3.3 (Nov 2019)
# Pipeline mode:   SCAN
# Query file:      AE014075.1_amino.fasta
# Target file:     Pfam-A.hmm
# Option settings: hmmscan --domtblout domtblout.txt --cut_ga Pfam-A.hmm AE014075.1_amino.fasta
# Current dir:     /Users/horta/code/iseq/scripts
# Date:            Wed Jul  8 17:21:50 2020
# [ok]
"""

_domtbl_empty_content = ""


def test_tbl():
    tbl = iter(read_tbl(StringIO(_tbl_content)))
    row = next(tbl)

    assert row.target.name == "item2"
    assert row.target.accession == "-"
    assert row.query.name == "Octapeptide"
    assert row.query.accession == "PF03373.14"
    assert row.full_sequence.e_value == "1.2e-07"
    assert row.full_sequence.score == "19.5"
    assert row.full_sequence.bias == "3.5"
    assert row.best_1_domain.e_value == "1.2e-07"
    assert row.best_1_domain.score == "19.5"
    assert row.best_1_domain.bias == "3.5"
    assert row.domain_numbers.exp == "1.0"
    assert row.domain_numbers.reg == 1
    assert row.domain_numbers.clu == 0
    assert row.domain_numbers.ov == 0
    assert row.domain_numbers.env == 1
    assert row.domain_numbers.dom == 1
    assert row.domain_numbers.rep == 1
    assert row.domain_numbers.inc == 1
    assert row.description == "Description one two three"

    row = next(tbl)

    assert row.target.name == "item3"
    assert row.target.accession == "-"
    assert row.query.name == "Octapeptide"
    assert row.query.accession == "PF03373.14"
    assert row.full_sequence.e_value == "1.2e-07"
    assert row.full_sequence.score == "19.5"
    assert row.full_sequence.bias == "3.5"
    assert row.best_1_domain.e_value == "1.2e-07"
    assert row.best_1_domain.score == "19.5"
    assert row.best_1_domain.bias == "3.5"
    assert row.domain_numbers.exp == "1.0"
    assert row.domain_numbers.reg == 1
    assert row.domain_numbers.clu == 0
    assert row.domain_numbers.ov == 0
    assert row.domain_numbers.env == 1
    assert row.domain_numbers.dom == 1
    assert row.domain_numbers.rep == 1
    assert row.domain_numbers.inc == 1
    assert row.description == "-"

    with pytest.raises(StopIteration):
        next(tbl)


def test_domtbl():
    tbl = iter(read_domtbl(StringIO(_domtbl_content)))
    row = next(tbl)

    assert row.target.name == "Leader_Thr"
    assert row.target.accession == "PF08254.12"
    assert row.target.length == 22
    assert row.query.name == "AE014075.1:190-252|amino|11"
    assert row.query.accession == "-"
    assert row.query.length == 21
    assert row.full_sequence.e_value == "5.3e-10"
    assert row.full_sequence.score == "38.8"
    assert row.full_sequence.bias == "17.5"
    assert row.domain.id == 1
    assert row.domain.size == 1
    assert row.domain.c_value == "3e-14"
    assert row.domain.i_value == "5.5e-10"
    assert row.domain.score == "38.8"
    assert row.domain.bias == "17.5"
    assert row.hmm_coord.start == 1
    assert row.hmm_coord.stop == 22
    assert row.ali_coord.start == 1
    assert row.ali_coord.stop == 21
    assert row.env_coord.start == 1
    assert row.env_coord.stop == 21
    assert row.acc == "0.99"
    assert row.description == "Threonine leader peptide"

    row = next(tbl)
    assert row.target.name == "Y1_Tnp"

    assert len(list(read_domtbl(StringIO(_domtbl_content)))) == 8


def test_domtbl_empty():
    tbl = iter(read_domtbl(StringIO(_domtbl_empty_content)))
    with pytest.raises(StopIteration):
        next(tbl)



from importlib import import_module as _import_module

from . import typing
from ._example import example_filepath
from ._testit import test
from .domtbl import read_domtbl
from .hmmer import HMMER
from .tbl import read_tbl

try:
    __version__ = getattr(_import_module("hmmer._version"), "version", "x.x.x")
except ModuleNotFoundError:
    __version__ = "x.x.x"

__all__ = [
    "HMMER",
    "__version__",
    "example_filepath",
    "read_domtbl",
    "read_tbl",
    "test",
    "typing",
]



import csv
import dataclasses
from dataclasses import dataclass
from pathlib import Path
from typing import IO, List, Union

from ._misc import decomment

__all__ = [
    "DomTBLCoord",
    "DomTBLDomScore",
    "DomTBLIndex",
    "DomTBLRow",
    "DomTBLSeqScore",
    "read_domtbl",
]


@dataclass
class DomTBLIndex:
    name: str
    accession: str
    length: int


@dataclass
class DomTBLSeqScore:
    e_value: str
    score: str
    bias: str


@dataclass
class DomTBLDomScore:
    id: int
    size: int
    c_value: str
    i_value: str
    score: str
    bias: str


@dataclass
class DomTBLCoord:
    start: int
    stop: int


@dataclass
class DomTBLRow:
    target: DomTBLIndex
    query: DomTBLIndex
    full_sequence: DomTBLSeqScore
    domain: DomTBLDomScore
    hmm_coord: DomTBLCoord
    ali_coord: DomTBLCoord
    env_coord: DomTBLCoord
    acc: str
    description: str

    def _asdict(self):
        return dataclasses.asdict(self)

    def __iter__(self):
        return iter(self._asdict().values())

    def _field_types(self):
        return {f.name: f.type for f in dataclasses.fields(self)}


def read_domtbl(file: Union[str, Path, IO[str]]) -> List[DomTBLRow]:
    """
    Read domtbl file type.

    Parameters
    ----------
    file
        File path or file stream.
    """
    closeit = False
    if isinstance(file, str):
        file = Path(file)

    if isinstance(file, Path):
        file = open(file, "r")
        closeit = True

    rows = []
    for line in csv.reader(decomment(file), delimiter=" ", skipinitialspace=True):
        row = DomTBLRow(
            DomTBLIndex(line[0], line[1], int(line[2])),
            DomTBLIndex(line[3], line[4], int(line[5])),
            DomTBLSeqScore(*line[6:9]),
            DomTBLDomScore(int(line[9]), int(line[10]), *line[11:15]),
            DomTBLCoord(int(line[15]), int(line[16])),
            DomTBLCoord(int(line[17]), int(line[18])),
            DomTBLCoord(int(line[19]), int(line[20])),
            line[21],
            " ".join(line[22:]),
        )
        rows.append(row)

    if closeit:
        file.close()

    return rows



from pathlib import Path
from typing import Union

__all__ = ["decomment", "make_path"]


def decomment(rows):
    for row in rows:
        if row.startswith("#"):
            continue
        yield row


def make_path(path: Union[Path, str]) -> Path:
    if isinstance(path, str):
        path = Path(path)
    return path



from .domtbl import DomTBLCoord, DomTBLDomScore, DomTBLIndex, DomTBLRow, DomTBLSeqScore
from .hmmer import Result
from .tbl import TBLDom, TBLIndex, TBLRow, TBLScore

__all__ = [
    "DomTBLCoord",
    "DomTBLDomScore",
    "DomTBLIndex",
    "DomTBLRow",
    "DomTBLSeqScore",
    "Result",
    "TBLDom",
    "TBLIndex",
    "TBLRow",
    "TBLScore",
]



from pathlib import Path as _Path
from sys import platform as _platform

__all__ = ["hmmfetch", "hmmpress", "hmmscan", "hmmsearch", "hmmemit"]

if _platform not in ["linux", "darwin"]:
    raise RuntimeError(f"Unsupported platform: {_platform}.")

_suffix = "manylinux2010_x86_64"
if _platform == "darwin":
    _suffix = "macosx_10_9_x86_64"


_bin = _Path(__file__).parent.absolute()

hmmemit = _bin / f"hmmemit_{_suffix}"
hmmfetch = _bin / f"hmmfetch_{_suffix}"
hmmpress = _bin / f"hmmpress_{_suffix}"
hmmscan = _bin / f"hmmscan_{_suffix}"
hmmsearch = _bin / f"hmmsearch_{_suffix}"
